#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/wait.h>
#include "iolib.h"
#define PROGNAME "servetcp"
#include "freedt.h"

int max_conns = 40;
int conn_count = 0;
int gid = -1, uid = -1;
int backlog = 10;
int kill_options = 1;
int no_delay = 0;
int verbose = 0;
char **command;
int listen_fd;
int child_pipe[2];

void handle_sigchld(int dummy) {
	int n;
	do {
		n = write(child_pipe[1], "c", 1);
	} while (n < 0 && errno == EINTR);
	if (n < 1)
		die("unable to write to pipe");
}

void fd_set_add(int fd, fd_set *fds, int *max) {
	FD_SET(fd, fds);
	if (fd > *max) *max = fd;
}

int main(int argc, char **argv) {
	struct sigaction sa;
	struct sockaddr_in listen_addr;
	char *s;
	int n;
	
	istdinit();

	while (1) {
		char c = getopt(argc, argv, "+c:g:u:Ub:oOdDQv");
		if (c == -1)
			break;
		switch (c) {
		case 'c':
			max_conns = atoi(optarg);
			break;
		case 'g':
			gid = atoi(optarg);
			break;
		case 'u':
			uid = atoi(optarg);
			break;
		case 'U':
			s = getenv("GID");
			if (!s)
				die("-U specified but no $GID");
			gid = atoi(s);
			s = getenv("UID");
			if (!s)
				die("-U specified but no $UID");
			uid = atoi(s);
			break;
		case 'b':
			backlog = atoi(optarg);
			break;
		case 'o':
			kill_options = 0;
			break;
		case 'O':
			kill_options = 1;
			break;
		case 'd':
			no_delay = 0;
			break;
		case 'D':
			no_delay = 1;
			break;
		case 'Q':
			verbose = 0;
			break;
		case 'v':
			verbose = 1;
			break;
		default:
			die("unknown option");
		}
	}

	if ((argc - optind) < 3)
		die("usage: servetcp [-c maxconns] [-g gid] [-u uid] "
			"[-b backlog] [-UoOdDQv] host port command ...");

	if (pipe(child_pipe) < 0)
		die("unable to create pipe");

	listen_addr.sin_family = AF_INET;

	s = argv[optind++];
	if (strcmp(s, "0") == 0) {
		listen_addr.sin_addr.s_addr = INADDR_ANY;
	} else if (inet_aton(s, &listen_addr.sin_addr) != 0) {
		/* nothing */
	} else {
		struct hostent *he = gethostbyname(s);
		if ((!he) || (he->h_addrtype != AF_INET) || (he->h_addr == 0))
			die2(s, "unable to resolve listen host");
		listen_addr.sin_addr = *(struct in_addr *)he->h_addr;
	}

	s = argv[optind++];
	n = atoi(s);
	if (n == 0 && strcmp(s, "0") != 0) {
		struct servent *se = getservbyname(s, "tcp");
		if (!se)
			die("unable to resolve listen port");
		n = se->s_port;
	} else {
		n = htons(n);
	}
	listen_addr.sin_port = n;
	command = &argv[optind];

	listen_fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (listen_fd < 0)
		die("unable to create socket");
	if (fcntl(listen_fd, F_SETFL, O_NONBLOCK) < 0)
		die("unable to set O_NONBLOCK");
	n = 1;
	if (setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR, &n, sizeof n) < 0)
		die("unable to set SO_REUSEADDR");
	if (bind(listen_fd, (struct sockaddr *)&listen_addr,
		sizeof listen_addr) < 0)
		die("unable to bind to listen address");
	if (listen(listen_fd, backlog) < 0)
		die("unable to listen");

	if (gid != -1)
		if (setgid(gid) < 0)
			die("unable to setgid");
	if (uid != -1)
		if (setuid(uid) < 0)
			die("unable to setuid");

	sa.sa_handler = handle_sigchld;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_NOCLDSTOP;
	sigaction(SIGCHLD, &sa, NULL);

	while (1) {
		int max;
		fd_set read_fds;

		FD_ZERO(&read_fds);
		fd_set_add(child_pipe[0], &read_fds, &max);
		if (conn_count < max_conns)
			fd_set_add(listen_fd, &read_fds, &max);

		n = select(max + 1, &read_fds, NULL, NULL, NULL);
		if (n < 0 && errno != EINTR)
			die("select failed");

		if (FD_ISSET(child_pipe[0], &read_fds)) {
			char c;
			do {
				n = read(child_pipe[0], &c, 1);
			} while (n < 0 && errno == EINTR);
			if (n < 1)
				die("unable to read from pipe");
			n = waitpid(-1, NULL, WNOHANG);
			if (n > 0) {
				if (verbose)
					iformat(ierr, 'i', n,
						'c', " closed\n", 0);
				conn_count--;
			}
		}

		if (FD_ISSET(listen_fd, &read_fds)) {
			int pid;
			struct sockaddr_in child_addr;
			int len = sizeof child_addr;
			int child_fd = accept(listen_fd,
				(struct sockaddr *)&child_addr, &len);
			
			if (child_fd < 0 && (errno == EAGAIN || errno == EINTR))
				goto no_conn;
			if (child_fd < 0)
				die("accept failed");
			n = 1;
			if (no_delay && setsockopt(child_fd, IPPROTO_TCP,
				TCP_NODELAY, &n, sizeof n) < 0)
				die("unable to set TCP_NODELAY");

			n = 0;
			if (kill_options && setsockopt(child_fd, SOL_IP,
				IP_OPTIONS, &n, sizeof n) < 0)
				die("unable to kill IP options");

			conn_count++;
			pid = fork();
			if (pid < 0)
				die("fork failed");

			if (pid == 0) {
				buffer b = BUFFER;

				close(listen_fd);
				close(child_pipe[0]);	
				close(child_pipe[1]);
				dup2(child_fd, 0);
				dup2(child_fd, 1);

				setenv("PROTO", "TCP", 1);
				bformat(&b, 'a', listen_addr.sin_addr.s_addr,
					0);
				setenv("TCPLOCALIP", bstr(&b), 1);
				bfree(&b);
				bformat(&b, 'i', ntohs(listen_addr.sin_port),
					0);
				setenv("TCPLOCALPORT", bstr(&b), 1);
				bfree(&b);
				bformat(&b, 'a', child_addr.sin_addr.s_addr,
					0);
				setenv("TCPREMOTEIP", bstr(&b), 1);
				bfree(&b);
				bformat(&b, 'i', ntohs(child_addr.sin_port),
					0);
				setenv("TCPREMOTEPORT", bstr(&b), 1);
				bfree(&b);

				execvp(command[0], command);
				_exit(20);
			}

			close(child_fd);
			if (verbose)
				iformat(ierr, 'i', pid, 'c', " connected from ",
					'a', child_addr.sin_addr.s_addr,
					'c', " port ",
					'i', ntohs(child_addr.sin_port),
					'c', "\n", 0);

			no_conn: ;
		}	
	}

	return 0;
}

